import * as React from 'react';
import BaseComponent from '../BaseComponent';
import Toolbar from './toolbar/index';
import UndoUtil from './utils/undoUtil';
import {MainContext} from './mainContext';
import './index.scss';
import '../../assets/symbol';

const IMAGE_TYPE = 'image/png';

type DoodleBoardProps = {
  width: number,
  height: number,
  initBoardImg?: string | null
  backdropImg?: string | null,
  conStrain?: Boolean
  showToolbars: boolean,
  toolbars: string[],
  onSave?: Function,
  onCancel?: Function,
}
type Result = {
  additionalImg: string,
  composeImg: string,
  composeImgWithBlog: Blob
}
export default class DoodleBoard extends BaseComponent {
  static defaultProps: DoodleBoardProps = {
    width: 500, // 画布宽度
    height: 500, // 画布高度
    backdropImg: null,  // 外部背景图
    conStrain: true, // 如果有背景，是否适应背景大小调整画布
    initBoardImg: null, // 初始画布数据，
    showToolbars: true, // 是否展示工具栏
    toolbars: [ // 功能列表
      'brush',  // 画笔
      'rectangle',// 正方形
      'undo',   // 撤销
      'redo',   // 重做
      'eraser', // 橡皮
      'cancel', // 取消按钮，如有，需要添加cancel方法
      'save',  // 保存按钮，如有，需要添加onSave方法
    ],
    onSave: (result: Result) => {
    }, // 点击
    onCancel: () => {
    }
  };
  // 背景canvas以及其context
  private backdropCanvasEle: null | HTMLCanvasElement;
  private backdropCanvasContext: null | CanvasRenderingContext2D;
  // 画板canvas以及其context
  private boardCanvasEle: null | HTMLCanvasElement;
  private boardCanvasContext: null | CanvasRenderingContext2D;
  // 是否正在绘图中
  private isDrawing: boolean;
  // 当前UI主要是draw方法绘图
  private crtUI: any;
  private undoUtil: UndoUtil;
  // 起始着笔点
  private startPoint: { x: number, y: number } = {x: 0, y: 0};

  constructor(props: any) {
    super(props);
    this.backdropCanvasEle = null;
    this.backdropCanvasContext = null;
    this.boardCanvasEle = null;
    this.boardCanvasContext = null;
    this.isDrawing = false;
    this.crtUI = null;
    // 默认canvasConfig,
    this.state = {
      defaultCanvasConfig: {
        lineWidth: 1,
        lineColor: '#ff0200',
        shadowBlur: 1
      },
      canvasWidth: this.props.width,
      canvasHeight: this.props.height,
    };
    // 设置当前UI
    this.setUI = this.setUI.bind(this);
    this._onMouseDown = this._onMouseDown.bind(this);
    this._onMouseUp = this._onMouseUp.bind(this);
    this._onMouseMove = this._onMouseMove.bind(this);
    this._refBoardCanvas = this._refBoardCanvas.bind(this);
    this._refBackdropCanvas = this._refBackdropCanvas.bind(this);
  }

  componentDidMount() {
    // 渲染背景
    this._renderBackDrop();
    this._renderInitBoard();
    // this._initDrawEvents();
  }

  _refBackdropCanvas(ele: HTMLCanvasElement | null) {
    this.backdropCanvasEle = ele;
    this.backdropCanvasContext = !!ele ? ele.getContext('2d') : null;
  }

  _refBoardCanvas(ele: HTMLCanvasElement | null) {
    this.boardCanvasEle = ele;
    this.boardCanvasContext = !!ele ? ele.getContext('2d') : null;
    // 实例化undoutil
    // if (!this.undoUtil) {
    //   this.undoUtil = new UndoUtil(this._getCrtCanvasImageData());
    // }
  }

  // 绘图事件
  _initDrawEvents() {
    if (!!this.boardCanvasEle) {
      // this.boardCanvasEle.addEventListener('mousedown', this._onMouseDown);
      // this.boardCanvasEle.addEventListener('mouseup', this._onMouseUp);
      // this.boardCanvasEle.addEventListener('mousemove', this._onMouseMove);
    }

  }

  // 组件移除时销毁事件
  componentWillUnmount() {
    if (!!this.boardCanvasEle) {
      // this.boardCanvasEle.removeEventListener('mousedown', this._onMouseDown);
      // this.boardCanvasEle.removeEventListener('mouseup', this._onMouseUp);
      // this.boardCanvasEle.removeEventListener('mousemove', this._onMouseMove);
    }

  }

  /**
   * 按钮按下，表明涂鸦已经开始，记录一下当前图像，用于撤销
   * @param e
   * @private
   */
  _onMouseDown = (e: any) => {
    console.log('mouse down');
    this.isDrawing = true;
    this._setCanvasStyle();
    const {offsetX, offsetY} = e.nativeEvent;
    if (!!this.boardCanvasContext) {
      this.boardCanvasContext.globalCompositeOperation = 'source-over';
      this.boardCanvasContext.beginPath();
      this.boardCanvasContext.moveTo(offsetX, offsetY);
    }
    this.startPoint = {
      x: offsetX,
      y: offsetY
    }
  };

  /**
   * 鼠标弹起，表明涂鸦结束
   * @private
   */
  _onMouseUp = () => {
    console.log('mouse up');
    this.isDrawing = false;
    this.undoUtil.save(this._getCrtCanvasImageData());
  };

  /**
   * 鼠标移动，绘图
   * @param e
   * @private
   */
  _onMouseMove = (e: any) => {
    if (this.isDrawing) {
      console.log('mouse move');
      this.crtUI.draw(this.boardCanvasEle, this.startPoint, e.nativeEvent);
    }
  };

  getLatestImageData() {
    if (!!this.undoUtil) {
      return this.undoUtil.current();
    }
    return null;
  }

  /**
   * @return {ImageData}
   * @private
   */
  _getCrtCanvasImageData() {
    return !!this.boardCanvasContext ? this.boardCanvasContext.getImageData(0, 0, this.props.width, this.props.height) : null;
  }

  _setCanvasStyle() {
    if (!!this.boardCanvasContext) {
      this.boardCanvasContext.lineWidth = this.state.defaultCanvasConfig.lineWidth;
      this.boardCanvasContext.shadowBlur = this.state.defaultCanvasConfig.shadowBlur;
      this.boardCanvasContext.shadowColor = this.state.defaultCanvasConfig.lineColor;
      this.boardCanvasContext.strokeStyle = this.state.defaultCanvasConfig.lineColor;
    }
  }

  /**
   * 获取画布的大小， 主要是有背景图且设置conStrain为true
   * 这样画布大小会根据背景图动态调整,同时width以及height将升级成最大宽高
   * @private
   */
  _calcCanvasSize(initialCanvasWidth: number, initialCanvasHeight: number, imgWidth: number, imgHeight: number)
  : { canvasWidth: number, canvasHeight: number } {
    // 如果宽度高度都大于设置width,height,宽为最大，高度根据比例调整
    // width/height = iw/ih => iw*height = width*ih   =>(width = iw*height/ih) => (height = width*ih/iw)
    let canvasWidth : number = 0;
    let canvasHeight : number = 0;
    
    if (imgWidth > initialCanvasWidth) {
      canvasWidth = initialCanvasWidth;
      canvasHeight = initialCanvasWidth * imgHeight / imgWidth;
      
      // 如果计算后的高度仍然超出初始画布高度（例如图片尺寸为3024x4032），则还要继续调整图片的比例
      if (canvasHeight > initialCanvasHeight) {
        return this._calcCanvasSize(initialCanvasWidth, initialCanvasHeight, canvasWidth, canvasHeight );
      }
    // 如果宽度小但是高度比height高
    } else if (imgHeight > initialCanvasHeight) {
      canvasWidth = imgWidth * initialCanvasHeight / imgHeight;
      canvasHeight = initialCanvasHeight
    } else {
      canvasWidth = imgWidth;
      canvasHeight = imgHeight;
    }
    
    // 如果宽高都小，按照原图宽高来
    return {
      canvasWidth,
      canvasHeight
    }
  } 
  
  _getCanvasSize(): Promise<{ width: number, height: number }> {
    const {backdropImg, conStrain, width, height} = this.props;
    if (!!backdropImg && conStrain) {
      const img = new Image();
      img.setAttribute('crossOrigin', 'anonymous');
      img.src = backdropImg;
      return new Promise((resolve) => {
        img.onload = () => {
          const { canvasWidth, canvasHeight } = this._calcCanvasSize(width, height, img.width, img.height);
          
          this.setState({
            canvasWidth,
            canvasHeight
          });
          resolve({
            width: canvasWidth,
            height: canvasHeight
          })
        }
      })
    } else {
      this.setState({
        canvasWidth: width,
        canvasHeight: height
      });
      return Promise.resolve({
        width,
        height
      })
    }
  }

  /**
   * 渲染背景图 如果有
   * @private
   */
  _renderBackDrop() {
    // const {backdropImg, width, height} = this.props;
    const {backdropImg} = this.props;
    if (!!backdropImg) {
      const img = new Image();
      img.setAttribute('crossOrigin', 'anonymous');
      img.src = backdropImg;
      img.onload = async () => {
        if (!!this.backdropCanvasContext) {
          const canvasSize = await this._getCanvasSize();
          this.backdropCanvasContext.clearRect(0, 0, canvasSize.width, canvasSize.height);
          this.backdropCanvasContext.drawImage(img, 0, 0, canvasSize.width, canvasSize.height);
        }
      };
    }
  }

  _renderInitBoard() {
    // const {initBoardImg, width, height} = this.props;
    const {initBoardImg} = this.props;
    if (!!initBoardImg) {
      const img = new Image();
      img.setAttribute('crossOrigin', 'anonymous');
      img.src = initBoardImg;
      img.onload = async () => {
        if (!!this.boardCanvasContext) {
          const canvasSize = await this._getCanvasSize();
          this.boardCanvasContext.clearRect(0, 0, canvasSize.width, canvasSize.height);
          this.boardCanvasContext.drawImage(img, 0, 0, canvasSize.width, canvasSize.height);
          // 如果有初始数据，那么初始算作第一个Undo
          this.undoUtil = new UndoUtil(this._getCrtCanvasImageData());
        }
      };
    } else {
      this.undoUtil = new UndoUtil(this._getCrtCanvasImageData());
    }
  }


  setUI(ui: any) {
    this.crtUI = ui;
  }

  /**
   * 撤销
   */
  undo() {
    const imgData = this.undoUtil.undo();
    if (!!imgData && !!this.boardCanvasContext) {
      this.boardCanvasContext.clearRect(0, 0, this.props.width, this.props.height);
      this.boardCanvasContext.putImageData(imgData, 0, 0);
    }
  }

  /**
   * 重做
   */
  redo() {
    const imgData = this.undoUtil.redo();
    if (!!imgData && !!this.boardCanvasContext) {
      this.boardCanvasContext.clearRect(0, 0, this.props.width, this.props.height);
      this.boardCanvasContext.putImageData(imgData, 0, 0);
    }
  }

  async save() {
    if (!!this.props.onSave) {
      const blobData = await this.generateImgWithBlob();
      this.props.onSave({
        additionalImg: this.generateAdditionalImgWithDataUrl(),
        composeImg: this.generateImgWithDataUrl(),
        composeImgWithBlog: blobData
      })
    }
  }

  cancel() {
    if (!!this.props.onCancel) {
      this.props.onCancel();
    }
  }

  /**
   * 生成绘图板数据
   */
  generateAdditionalImgWithDataUrl(): string | null {
    if (!!this.boardCanvasEle) {
      return this.boardCanvasEle.toDataURL(IMAGE_TYPE);
    }
    return null;
  }

  /**
   * @return Promise<blob|null>
   * 生成blob类型的图像，如果没有画板实例，返回null
   */
  async generateImgWithBlob(): Promise<Blob | null> {
    const resultCanvasEle = document.createElement('canvas');
    const ctx = resultCanvasEle.getContext('2d');
    resultCanvasEle.width = this.state.canvasWidth;
    resultCanvasEle.height = this.state.canvasHeight;
    if (!!ctx && !!this.backdropCanvasEle && !!this.boardCanvasEle) {
      ctx.drawImage(this.backdropCanvasEle, 0, 0);
      ctx.drawImage(this.boardCanvasEle, 0, 0);
      return new Promise((resolve) => {
        resultCanvasEle.toBlob(function (result) {
          resolve(result);
        }, IMAGE_TYPE, 1);
      });
    }
    return Promise.resolve(null)
  }

  /**
   * 生成base64图像
   */
  generateImgWithDataUrl(): string | null {
    const resultCanvasEle = document.createElement('canvas');
    const ctx = resultCanvasEle.getContext('2d');
    resultCanvasEle.width = this.state.canvasWidth;
    resultCanvasEle.height = this.state.canvasHeight;
    if (!!ctx && !!this.backdropCanvasEle && !!this.boardCanvasEle) {
      ctx.drawImage(this.backdropCanvasEle, 0, 0);
      ctx.drawImage(this.boardCanvasEle, 0, 0);
      return resultCanvasEle.toDataURL(IMAGE_TYPE);
    }
    return null
  }


  render() {
    let width = `${this.props.width}px`;
    let height = `${this.props.height}px`;
    const {toolbars, showToolbars} = this.props;
    let canvasWidth = `${this.state.canvasWidth}px`;
    let canvasHeight = `${this.state.canvasHeight}px`;
    return (
      <MainContext.Provider value={{mainContext: this}}>
        <div className="bs-doodle-board-container" style={{width: width, height: height}}>
          <div className="bs-doodle-board-toolbar-container" style={{display: showToolbars ? 'block' : 'none'}}>
            <Toolbar toolbars={toolbars} setUI={this.setUI}/>
          </div>
          <canvas className="bs-doodle-board__canvas_back" width={canvasWidth} height={canvasHeight}
                  ref={this._refBackdropCanvas}>
          </canvas>
          <canvas className="bs-doodle-board__canvas_board"
                  width={canvasWidth} height={canvasHeight} ref={this._refBoardCanvas}
                  onMouseDown={this._onMouseDown}
                  onMouseUp={this._onMouseUp}
                  onMouseMove={this._onMouseMove}
          >
          </canvas>
        </div>
      </MainContext.Provider>
    );
  }
}
